@@ -8,17 +8,17 @@ use warnings;
 # no Carp::Assert qw(assert DEBUG);
 
 our $VERSION
- = '1.36';
+ = '1.42';
 
 use enum qw( HEADER=0 KEY VALUE );
 
 sub new {
   my ($class, $key, $value, $hdr) = @_;
-  my $self  = [ ];
+  my $self  = [ ($hdr || [ ]), $key, $value ];
 
-  $self->[KEY]    = $key;
-  $self->[VALUE]  = $value;
-  $self->[HEADER] = $hdr || [ ];
+#   $self->[KEY]    = $key;
+#   $self->[VALUE]  = $value;
+#   $self->[HEADER] = $hdr || [ ];
 
   bless $self, $class;
 }
@@ -29,7 +29,7 @@ sub header {
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 #   assert( (!$hdr) || ref($hdr) eq 'ARRAY' ), if DEBUG;
 
-  ($hdr) ? ( $self->[HEADER] = $hdr ) : $self->[HEADER];
+  $self->[HEADER];
 }
 
 # sub prev {
@@ -57,7 +57,7 @@ sub key {
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 #   assert( (@_==1) || $self->validate_key( $key ) ), if DEBUG;
 
-  (@_ > 1) ? ( $self->[KEY] = $key ) : $self->[KEY];
+  $self->[KEY];
 }
 
 sub key_cmp {
@@ -87,6 +87,7 @@ sub value {
   (@_ > 1) ? ( $self->[VALUE] = $value ) : $self->[VALUE];
 }
 
+
 1;
 
 package List::SkipList::Header;
@@ -125,7 +126,7 @@ use warnings;
 # use Carp::Assert qw( assert DEBUG );
 
 our $VERSION
- = '0.04';
+ = '0.05';
 
 our @ISA = qw( List::SkipList::Node );
 
@@ -143,27 +144,62 @@ sub key_cmp {
   1;   # Note that the header returns "1" instead of "-1"!
 }
 
+sub key {
+#    assert( 0 ), if DEBUG;
+  return;
+}
+
+sub value {
+#    assert( 0 ), if DEBUG;
+  return;
+}
+
 1;
 
 package List::SkipList;
 
 use 5.006;
 use strict;
-use warnings;
+use warnings::register __PACKAGE__;
 
-our $VERSION = '0.65';
+our $VERSION = '0.71';
+$VERSION = eval $VERSION;
 
 use AutoLoader qw( AUTOLOAD );
 use Carp qw( carp croak );
 # no Carp::Assert qw(assert DEBUG);
 
+# Future versions should check Config module to determine if it is
+# being run on a 64-bit processor, and set MAX_LEVEL to 64.
+
+use constant MIN_LEVEL       =>  2;
 use constant MAX_LEVEL       => 32;
 use constant DEF_P           => 0.25;
+use constant DEF_K           => 0;
 
 use constant BASE_NODE_CLASS => 'List::SkipList::Node';
 
-my $NULL; INIT { $NULL = new List::SkipList::Null(); }
+# $NULL is a workaround for situations where a reference might be
+# 'undef' (e.g. $fwd = $node->header->[0] || $NULL), and will probably
+# be phased out of future versions.   We still use 'undef' as an
+# indicator that we're at the end of the list.
 
+our $NULL =  new List::SkipList::Null();
+
+# We use Exporter instead of something like Exporter::Lite because
+# Carp uses it.
+
+require Exporter;
+
+our @EXPORT    = ( );
+our @EXPORT_OK = ( );
+
+sub import {
+  unless ($NULL) {
+    $NULL = new List::SkipList::Null();
+  }
+  goto &Exporter::import;
+}
 
 sub new {
   no integer;
@@ -179,15 +215,18 @@ sub new {
     SIZE_LEVEL     => undef,            # maximum level random_level
     MAXLEVEL  => MAX_LEVEL,             # absolute maximum level
     P         => 0,                     # probability for each level
+    K         => 0,                     # minimum power of P
     P_LEVELS  => [ ],                   # array used by random_level
-    LASTNODE  => undef,                 # node with greatest key
+    LIST_END  => undef,                 # node with greatest key
     LASTKEY   => undef,                 # last key used by next_key
     LASTINSRT => undef,                 # cached insertion fingers
+    DUPLICATES => 0,                    # allow duplicates?
   };
 
   bless $self, $class;
 
   $self->_set_p( DEF_P ); # initializes P_LEVELS
+  $self->_set_k( DEF_K );
 
   if (@_) {
     my %args = @_;
@@ -206,6 +245,12 @@ sub new {
   return $self;
 }
 
+sub _set_duplicates {
+  my ($self, $dup) = @_;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+  $self->{DUPLICATES} = $dup || 0;
+}
+
 sub _set_node_class {
   my ($self, $node_class) = @_;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
@@ -232,16 +277,17 @@ sub clear {
   $self->{SIZE}     = 0;
   $self->{SIZE_THRESHOLD} = 2;
   $self->{LAST_SIZE_TH}   = 0;
-  $self->{SIZE_LEVEL}     = 2;
+  $self->{SIZE_LEVEL}     = MIN_LEVEL;
 
   my $hdr = [ (undef) x $self->{SIZE_LEVEL} ];
 
+  CORE::delete $self->{LIST};
   $self->{LIST} = new List::SkipList::Header( undef, undef, $hdr );
 
 #   assert( $self->list->level > 0 ), if DEBUG;
 #   assert( UNIVERSAL::isa($self->{LIST}, BASE_NODE_CLASS) ), if DEBUG;
 
-  $self->{LASTNODE}  = undef;
+  $self->{LIST_END}  = undef;
   $self->{LASTINSRT} = undef;
 
   $self->reset;
@@ -251,49 +297,98 @@ sub _set_max_level {
   my ($self, $level) = @_;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 #   assert( ($level>1) ), if DEBUG;
+  if ($level > MAX_LEVEL) {
+    croak "Cannot set max_level greater than ", MAX_LEVEL;
+  } elsif ($level < MIN_LEVEL) {
+    croak "Cannot set max_level less than ", MIN_LEVEL;
+  } elsif ((defined $self->list) && ($level < $self->list->level)) {
+    croak "Current level exceeds specified level";
+  }
   $self->{MAXLEVEL} = $level;
 }
 
 sub max_level {
-  my ($self) = @_;
+  my ($self, $level) = @_;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
-  $self->{MAXLEVEL};
+  if (defined $level) {
+    $self->_set_max_level($level);
+  } else {
+    $self->{MAXLEVEL};
+  }
+}
+
+# We use the formula from Pugh's "Skip List Cookbook" paper.  We
+# generate a reverse-sorted array of values based on p and k.  In
+# _new_node_level() we look for the highest value in the array that is
+# less than a random number n (0<n<1).
+
+sub _build_distribution {
+  no integer;
+
+  my ($self) = @_;
+
+  my $p = $self->p;
+  my $k = $self->k;
+
+  $self->{P_LEVELS} = [ (0) x MAX_LEVEL ]; 
+  for my $i (0..MAX_LEVEL) {
+    $self->{P_LEVELS}->[$i] = $p**($i+$k);
+  }
 }
 
 sub _set_p {
   no integer;
 
   my ($self, $p) = @_;
-#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
-  
+
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;  
+
+  unless ( ($p>0) && ($p<1) ) {
+    croak "Unvalid value for P (must be between 0 and 1)";
+  }
+
 #   assert( ($p>0) && ($p<1) ), if DEBUG;
 
   $self->{P} = $p;
+  $self->_build_distribution;
 
-  # Because configuration is via hash, we may not set a new max_level
-  # before setting P, so we set to the standard MAX_LEVEL if it is
-  # greater.  Possible bug is if max_level is greater than MAX_LEVEL
-  # but is set after P.
-
-  my $n     = 1;
-  my $level = 0;
-  my $max   = (MAX_LEVEL > $self->max_level) ? MAX_LEVEL : $self->max_level;
-
-  $self->{P_LEVELS} = [ ]; 
-
-  while ($level <= $max) {
-    # TODO: add assertion that [level]<[level-1]
-    $self->{P_LEVELS}->[$level++] = $n;
-    $n *= $p;
-  }
 }
 
 sub p {
   no integer;
 
-  my ($self) = @_;
+  my ($self, $p) = @_;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
-  $self->{P};
+
+  if (defined $p) {
+    $self->_set_p($p);
+  } else {
+    $self->{P};
+  }
+}
+
+sub _set_k {
+  my ($self, $k) = @_;
+
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;  
+
+  unless ( $k>=0 ) {
+    croak "Unvalid value for K (must be at least 0)";
+  }
+
+  $self->{K} = $k;
+  $self->_build_distribution;
+}
+
+sub k {
+  my ($self, $k) = @_;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+
+  if (defined $k) {
+    $self->_set_k($k);
+  } else {
+    $self->{K};
+  }
 }
 
 sub size {
@@ -308,27 +403,13 @@ sub list {
   $self->{LIST};
 }
 
-sub level {
-  my ($self) = @_;
-#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
-  $self->{LIST}->level;
-}
 
-sub null {
-#   my ($self) = @_;
-#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
-  $NULL;
-}
-
-sub _new_node_level { # previously _random_level
-  no integer;
+sub _adjust_level_threshold {
+  use integer;
 
   my ($self) = @_;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
-  # Note: we call $self->{MAXLEVEL} instead of $self->max_level for
-  # very minor speed improvement.
-
   if ($self->{SIZE} >= $self->{SIZE_THRESHOLD}) {
     $self->{LAST_SIZE_TH}    = $self->{SIZE_THRESHOLD};
     $self->{SIZE_THRESHOLD} += $self->{SIZE_THRESHOLD};
@@ -336,16 +417,25 @@ sub _new_node_level { # previously _random_level
   } elsif ($self->{SIZE} < $self->{LAST_SIZE_TH}) {
     $self->{SIZE_THRESHOLD}  = $self->{LAST_SIZE_TH};
     $self->{LAST_SIZE_TH}    = $self->{LAST_SIZE_TH} / 2;
-    $self->{SIZE_LEVEL}--;
+    $self->{SIZE_LEVEL}--, if ($self->{SIZE_LEVEL} > MIN_LEVEL);
 #     assert( $self->{LAST_SIZE_TH} == int($self->{LAST_SIZE_TH}) ), if DEBUG;
 #     assert( $self->{SIZE_LEVEL} >= 1 ), if DEBUG;
   }
+}
+
+sub _new_node_level { # previously _random_level
+  no integer;
+
+  my ($self) = @_;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
   my $n     = CORE::rand();
   my $level = 1;
 
-  do { } while (($n < $self->{P_LEVELS}->[$level]) &&
-	 ($level++ < $self->{SIZE_LEVEL}));
+  while (($n < $self->{P_LEVELS}->[$level]) &&
+	 ($level < $self->{SIZE_LEVEL})) {
+    $level++;
+  }
 
 #   assert( ($level >= 1) && ($level <= $self->{SIZE_LEVEL}) ), if DEBUG;
   $level;
@@ -389,10 +479,15 @@ sub _search_with_finger {
       $node = $fwd;
     }
     $finger->[$level] = $node;
-  } while ((--$level>=0) && ($cmp));
+  } while ((--$level>=0)); # && ($cmp));
 
   $node = $fwd, unless ($cmp);
-  
+
+  # Ideally we could stop when $cmp == 0, but the update vector would
+  # not be complete for levels below $level.  insert still works, but
+  # delete and truncate have problems and need kluges to make up for
+  # that.
+
   #   assert( UNIVERSAL::isa($node, BASE_NODE_CLASS) ), if DEBUG;
 
   ($node, $finger, $cmp);
@@ -431,7 +526,7 @@ sub _search {
     }
   } while ((--$level>=0) && ($cmp));
 
-  $node = $fwd, unless ($cmp);
+  $node = $fwd; # , unless ($cmp); # Devel::Cover says it's never false
 
   #   assert( UNIVERSAL::isa($node, BASE_NODE_CLASS) ), if DEBUG;
 
@@ -462,7 +557,7 @@ sub insert {
 
   ($node, $finger, $cmp) = $self->_search_with_finger($key, $finger);
 
-  if ($cmp) {
+  if ($cmp || $self->{DUPLICATES}) {
 
     my $new_level = $self->_new_node_level;
 
@@ -478,17 +573,11 @@ sub insert {
     }
 
 
-#     my $next = $node_hdr->[0];
-#     if ($next) {
-#       $node->prev( $next->prev );
-#       $next->prev( $node );
-#     } else {
-#       $self->{LASTNODE} = $node;
-#     }
-
-    $self->{LASTNODE} = $node, unless ($node_hdr->[0]);
+    # We no longer set the LIST_END value, since it is the job of the
+    # _greatest_node method to find it, as needed.
 
     $self->{SIZE}++;
+    $self->_adjust_level_threshold;
   } else {
     $node->value($value);
   }
@@ -511,44 +600,38 @@ sub delete {
 
   my $list = $self->list;
 
-  my ($x, $update_ref, $cmp) = $self->_search_with_finger($key, $finger);
+  my ($node, $update_ref, $cmp) = $self->_search_with_finger($key, $finger);
 
   if ($cmp == 0) {
-    my $value = $x->value;
+    my $value = $node->value;
 
-    my $level = $x->level-1; 
-#     assert($level < @{$update_ref}), if DEBUG;
-
-    for (my $i=$level; $i>=0; $i--) {
-
-      my $y   = $update_ref->[$i] || $list;
-
-      # The top level of the finger points to the current node. The
-      # lower levels should be set to that node if they point to the
-      # start of the list.
-
-      if ($y == $list) { $y = $update_ref->[$level]; }
-
-      while ((my $fwd=$y->header()->[$i]) != $x) {
-	$y = $fwd;
-# 	assert( UNIVERSAL::isa($y, BASE_NODE_CLASS) ), if DEBUG;
-      }
-      $y->header()->[$i] = $x->header()->[$i];
+    # Note: it might make better sense to set $self->{LIST_END} = undef, and
+    # let the _greatest_node method search for it if it's needed again.
 
+    if (($self->{LIST_END}) && ($node == $self->{LIST_END})) {
+      $self->{LIST_END}  = $update_ref->[0];
     }
 
-#     my $next = $x->header()->[0];
-#     if ($next) { $next->prev( $x->prev ); }
+    my $level = $node->level; 
+#     assert($level < @{$update_ref}), if DEBUG;
 
-    # There's probably a smarter way to handle this, but this is the
-    # safest way.
+    for (my $i=0; $i<$level; $i++) {
+      $update_ref->[$i]->header()->[$i] = $node->header()->[$i];
+    }
+
+    # There's probably a smarter way to handle the last insert and
+    # last key values, but this is the fastest, easiest, safest and
+    # most consistent way.
 
     $self->{LASTINSRT} = undef;
+    $self->reset;
 
     $self->{SIZE} --;
+    $self->_adjust_level_threshold;
 
-    # We shouldn't need to "undef $x" here. The Garbage Collector
-    # should hanldle that.
+    # We shouldn't need to "undef $node" here. The Garbage Collector
+    # should hanldle that (especially if there's a finger that points
+    # to it somewhere).
 
     # Note: It doesn't seem to be a wise idea to return a search
     # finger for deletions without further analysis
@@ -556,6 +639,7 @@ sub delete {
     $value;
 
   } else {
+    carp "key not found", if (warnings::enabled);
     return;
   }
 }
@@ -601,70 +685,90 @@ sub find {
   ($cmp == 0) ? $node->value : undef;
 }
 
-sub last_key {
-  my ($self, $key, $finger, $value) = @_;
+sub _first_node { # actually this is the second node
+  my $self = shift;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
-  if (defined $key) {
-    $self->{LASTKEY} = [ $key, $finger || [ ], $value ];
+  my $list = $self->list;
+  my $node = $list->header()->[0];
+}
+
+
+sub last_key {
+  my ($self, $node, $index) = @_;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+
+  if (@_ > 1) {
+    $self->{LASTKEY} = [ $node, $index ];
+    my $check = $index || 0;
+    if (($check < 0) || ($check >= $self->size)) {
+      carp "index out of bounds", if (warnings::enabled);
+    }
+  }
+  else {
+    unless ($self->{LASTKEY}) {
+      $self->{LASTKEY} = [ $self->_first_node, 0 ];
+    }
+    ($node, $index) = @{ $self->{LASTKEY} };
   }
 
-  if (defined $self->{LASTKEY}) {
+  if ($node) {
     return (wantarray) ?
-      @{$self->{LASTKEY}} : $self->{LASTKEY}->[0];
+      ( $node->key, [ $node ], $node->value, $index ) : $node->key;
   } else {
     return;
   }
-
 }
 
 sub first_key {
   my $self = shift;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
-  my $list = $self->list;
-  my $fwd  = $list->header()->[0];
-  if ($fwd) {
-    return $self->last_key( $fwd->key, [( $list ) x $list->level],
-			    $fwd->value );
-  } else {
+  my $node = $self->_first_node;
+
+  if ($node) {
+    return $self->last_key( $node, 0);
+  }
+  else {
+    carp "no _first_node", if (warnings::enabled);
     return;
   }
 }
 
-
 sub next_key {
-  my $self = shift;
+  my ($self, $last_key, $finger) = @_;
+
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
-
-  my $last_key = shift;
-  my $finger   = shift;
-
-  unless (defined $last_key) {
-    ($last_key, $finger) = $self->last_key;
-  }
-
 #   if ($finger) {
 #     assert( ref($finger) eq 'ARRAY' ), if DEBUG;
 #   }
 
-  if (defined $last_key) {
-    my ($list, $update_ref, $cmp) =
-      $self->_search_with_finger($last_key, $finger);
+  my ($node, $cmp, $value, $index);
 
-    my $fwd  = $list->header()->[0];
- 
-    if ($cmp == 0) {
-      if (defined $fwd) {
-	$self->last_key($fwd->key, $update_ref, $fwd->value);
-      } else {
-	return;
-      }
-    } else {
+  if (defined $last_key) {
+    ($node, $finger, $cmp) = $self->_search_with_finger($last_key, $finger);
+
+    if ($cmp) {
+      carp "cannot find last_key", if (warnings::enabled);
       return;
     }
-  } else {
-    $self->first_key;
+  }
+  else {
+    ($node, $index) = @{ $self->{LASTKEY} || [ ] };
+    unless ($node) {
+      return $self->first_key;
+    }
+  }
+
+  if ($node) {
+    $node = $node->header()->[0];
+    return $self->last_key(
+      $node,
+      (($node && (defined $index)) ? ($index+1) : undef )
+    );
+  }
+  else {
+    return $self->reset;
   }
 }
 
@@ -689,28 +793,64 @@ BEGIN
 
 __END__
 
-sub _first_node { # actually this is the second node
-  my $self = shift;
+sub find_duplicates {
+  my ($self, $key, $finger) = @_;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
-  my $list = $self->list;
-  my $fwd  = $list->header()->[0];
-  if (defined $fwd) {
-    return ($fwd, scalar $list->header);
-  } else {
+#   if (defined $finger) {
+#     assert( UNIVERSAL::isa($finger, 'ARRAY') ), if DEBUG;
+#   }
+
+  my ($node, $update_ref, $cmp) = $self->_search_with_finger($key, $finger);
+
+  if ($cmp == 0) {
+    my @values = ( $node->value );
+
+    while ( ($node->header()->[0]) &&
+	    ($node->header()->[0]->key_cmp($key) == 0) ) {
+      $node = $node->header()->[0];
+      push @values, $node->value;
+    }
+
+    return @values;
+  }
+  else {
     return;
   }
 }
 
+sub level {
+  my $self = shift;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+  return $self->list->level;
+}
+
+sub _greatest_node {
+  my ($self) = @_;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+
+  my $list = $self->{LIST_END} || $self->list;
+
+  my $level = $list->level-1;
+  do {
+    while ($list->header()->[$level]) {
+      $list = $list->header()->[$level];
+    }
+  } while (--$level >=0);
+
+  $self->{LIST_END} = $list;
+}
+
 sub least {
   my $self = shift;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
-  my ($node, $finger) = $self->_first_node;
+  my ($node) = $self->_first_node;
 
-  if (defined $node) {
+  if ($node) {
     return ($node->key, $node->value);
   } else {
+    carp "no _first_node", if (warnings::enabled);
     return;
   }
 }
@@ -719,10 +859,11 @@ sub greatest {
   my $self = shift;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
-  my $node = $self->{LASTNODE};
-  if (defined $node) {
+  my $node = $self->_greatest_node;
+  if ($node) {
     return ($node->key, $node->value);
   } else {
+    carp "no _greatest_node", if (warnings::enabled);
     return;
   }
 }
@@ -759,11 +900,11 @@ sub keys {
 
   my @keys = ();
 
-  my ($key, $finger) = $self->first_key;
+  my $key = $self->first_key;
 
   while (defined $key) {
     push @keys, $key;
-    $key = $self->next_key($key, $finger);
+    $key = $self->next_key;
   }
 
   return @keys;
@@ -773,41 +914,139 @@ sub values {
   my $self = shift;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
+  my $key = $self->first_key;
+
   my @values = ();
 
-  my ($key, $finger) = $self->first_key;
-
   while (defined $key) {
-    push @values, scalar $self->find($key, $finger);
-    ($key, $finger) = $self->next_key($key, $finger);
+    my $value = ($self->last_key)[2];
+    push @values, $value;
+    $key = $self->next_key;
   }
 
   return @values;
 }
 
+sub truncate {
+  my $self = shift;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+
+  my ($key, $finger) = @_;
+
+  if (defined $key) {
+    my ($node, $finger, $cmp) = $self->_search_with_finger( $key, $finger );
+    if ($cmp == 0) {
+
+      # This is the most braindead way to find the index of a node. We
+      # could come up with more sophisticated way by saving the number
+      # of "skips" in the forward pointers when we add nodes, but that
+      # will significantly affect the speed.
+
+      my $size = 1 + $self->index_by_key( $key );
+#       {
+# 	my $aux  = $self->list;
+# 	while ($aux != $node) {
+# 	  $size++;
+# 	  $aux = $aux->header()->[0];
+# 	}
+#       }
+
+      my $list = new List::SkipList(
+        max_level  => $self->max_level,
+        p          => $self->p,
+        node_class => $self->_node_class,
+      );
+
+      my $level   = $self->list->level;
+      my $old_hdr = $self->list->header;
+      my $new_hdr = $list->list->header;
+
+      for (my $i=0; $i<$level; $i++) {
+
+	if ($finger->[$i]) {
+	  if ($finger->[$i] == $node) {
+	    $new_hdr->[$i] = $finger->[$i];
+	    $finger->[$i]  = undef;
+	  }
+	  else {
+	    $new_hdr->[$i] = $finger->[$i]->header()->[$i];
+	    $finger->[$i]->header()->[$i]  = undef;
+	  }
+	}
+	elsif ($old_hdr->[$i]) {
+
+	  if ($old_hdr->[$i] == $node) {
+	    $new_hdr->[$i] = $old_hdr->[$i];
+	    $old_hdr->[$i]  = undef;
+	  }
+	  else {
+	    carp "unexpected situation",
+	      if (warnings::enabled);
+	    # If _search_with_finger does not stop on !$cmp but
+	    # continues to remaining levels, then we should not
+	    # need to worry about this.
+	  }
+	}
+
+
+      }
+
+      $list->{SIZE} = $self->size - $size;
+      $self->{SIZE} = $size;
+
+      $list->{LIST_END} = undef;
+      $self->{LIST_END} = undef;
+
+      $self->_adjust_level_threshold;
+      $list->_adjust_level_threshold;
+
+      return $list;
+    }
+    else {
+    carp "key not found", if (warnings::enabled);
+      return;
+    }
+  }
+  else {
+    croak "no key specified";
+    return;
+  }
+
+}
+
+
 sub copy {
   my $self = shift;
 #   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
 
+  my ($key, $finger_or, $key_to) = @_;
+
   my $list = new List::SkipList(
     max_level  => $self->max_level,
     p          => $self->p,
     node_class => $self->_node_class,
   );
+  $list->{DUPLICATES} = $self->{DUPLICATES};
 
-  my ($key, $finger_or) = $self->first_key();
-  my $value;
+  my ($node, $cmp);
   my $finger_cp = undef;
 
-  $self->reset();
+  if (defined $key) {
+    ($node, $finger_or, $cmp) = $self->_search_with_finger( $key, $finger_or );
+    if ($cmp != 0) {
+      return;
+    }
+  }
+  else {
+    $node      = $self->_first_node;
+    $finger_or = undef;
+  }
 
-  do {
-    ($value, $finger_or) = $self->find_with_finger($key, $finger_or);
-    my $finger_cp        = $list->insert($key, $value, $finger_cp);
-    ($key, $finger_or)   = $self->next_key($key, $finger_or);
-  } while (defined $key);
-
-  $self->reset();
+  while ((defined $node) && 
+	 ((!defined $key_to) || ($node->key_cmp($key_to) <= 0))) {
+    my $finger_cp = $list->insert($node->key, $node->value, $finger_cp);
+    $node = $node->header()->[0];
+  }
 
   return $list;
 }
@@ -820,19 +1059,20 @@ sub merge {
   my $list2 = shift;
 #   assert( UNIVERSAL::isa($list2, __PACKAGE__) ), if DEBUG;
 
-  my ($node1, $finger1) = $list1->_first_node;
-  my ($node2, $finger2) = $list2->_first_node;
+  my ($finger1, $finger2);
+  my ($node1) = $list1->_first_node;
+  my ($node2) = $list2->_first_node;
 
 #   assert( ref($node1) eq ref($node2) ), if DEBUG;
 #   assert( ref($finger1) eq 'ARRAY' ), if DEBUG;
 
-  while ((defined $node1) || (defined $node2)) {
+  while ($node1 || $node2) {
 
-    my $cmp = (defined $node1) ? (
-     (defined $node2) ? $node1->key_cmp( $node2->key ) : 1 ) : -1;
+    my $cmp = ($node1) ? (
+      ($node2) ? $node1->key_cmp( $node2->key ) : 1 ) : -1;
     
     if ($cmp < 0) {                     # key1 < key2
-      if (defined $node1) {
+      if ($node1) {
 	$finger1 = $list1->insert( $node1->key, $node1->value, );
 	$node1 = $node1->header()->[0];
       } else {
@@ -840,7 +1080,7 @@ sub merge {
 	$node2 = $node2->header()->[0];
       }
     } elsif ($cmp > 0) {                # key1 > key2
-      if (defined $node2) {
+      if ($node2) {
 	$finger1 = $list1->insert( $node2->key, $node2->value, );
 	$node2 = $node2->header()->[0];
       } else {
@@ -849,9 +1089,9 @@ sub merge {
       }
     } else {                            # key1 = key2
       $node1 = $node1->header()->[0],
-	if defined $node1;
+	if $node1;
       $node2 = $node2->header()->[0],
-	if defined $node2;
+	if $node2;
     }
   }
 }
@@ -865,22 +1105,22 @@ sub append {
   unless (defined $list2) { return; }
 #   assert( UNIVERSAL::isa($list2, __PACKAGE__) ), if DEBUG;
 
-  my $node = $list1->{LASTNODE};
-  if (defined $node) {
+  my $node = $list1->_greatest_node;
+  if ($node) {
 
-    my ($next, $finger) = $list2->_first_node;
+    my ($next) = $list2->_first_node;
 
 #    assert( $node->key_cmp( $next->key ) < 0 ), if DEBUG;
 
-    if ($list1->level > $list2->level) {
+    if ($list1->list->level > $list2->list->level) {
 
-      if ($list1->level < $list1->max_level) {
+      if ($list1->list->level < $list1->max_level) {
 
-	my $i = $list1->level;
+	my $i = $list1->list->level;
 	while (!defined $list1->list->header()->[$i]) { $i--; }
 	$list1->list->header()->[$i+1] = $next;
       } else {
-	my $i = $list1->level-1;
+	my $i = $list1->list->level-1;
 	my $x = $list1->list->header()->[$i];
 	while (defined $x->header()->[$i]) {
 	  $x = $x->header()->[$i];
@@ -893,22 +1133,109 @@ sub append {
       for (my $i=0; $i<$node->level; $i++) {
 	$node->header()->[$i] = $next;
       }
-      for (my $i=$list1->level; $i<$list2->level; $i++) {
+      for (my $i=$list1->list->level; $i<$list2->list->level; $i++) {
 	$list1->list->header()->[$i] = $next;
       }
     }
 
     $list1->{SIZE}    += $list2->size;
-    $list1->{LASTNODE} = $list2->{LASTNODE};
-
+    $list1->{LIST_END} = $list2->{LIST_END};
   } else {
     $list1->{LIST}     = $list2->list;
     $list1->{SIZE}     = $list2->size;
-    $list1->{LASTNODE} = $list2->{LASTNODE};
+    $list1->{LIST_END} = $list2->{LIST_END};
+  }
+  $list1->_adjust_level_threshold;
+}
+
+sub _node_by_index {
+  my ($self, $index) = @_;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+
+  # Bug: for some reason, change $[ does not affect this module.
+
+#   if ($index >= $[) {
+#     $index -= $[;
+#   }
+
+  if ($index < 0) {
+    $index += $self->size;
   }
 
+  if (($index < 0) || ($index >= $self->size)) {
+    carp "index out of range", if (warnings::enabled);
+    return;
+  }
+
+
+  my ($node, $last_index) = @{ $self->{LASTKEY} || [ ] };
+
+  if ((defined $last_index)  && ($last_index <= $index)) {
+    ($last_index, $index) = ($index, $index - $last_index);
+  }
+  else {
+    $last_index = $index;
+    $node = undef;
+  }
+
+  $node ||= $self->_first_node;
+
+  unless ($node) {
+    return;
+  }
+
+  while ($node && $index--) {
+    $node = $node->header()->[0];
+  }
+
+  $self->last_key( $node, $last_index );
+  return $node;
 }
 
+sub key_by_index {
+  my ($self, $index) = @_;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+
+  my $node = $self->_node_by_index($index);
+  if ($node) {
+    return $node->key;
+  } else {
+    return;
+  }
+}
+
+sub value_by_index {
+  my ($self, $index) = @_;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+
+  my $node = $self->_node_by_index($index);
+  if ($node) {
+    return $node->value;
+  } else {
+    return;
+  }
+}
+
+sub index_by_key {
+  my ($self, $key) = @_;
+#   assert( UNIVERSAL::isa($self, __PACKAGE__) ), if DEBUG;
+
+  my $node  = $self->_first_node;
+  my $index = 0;
+  while ($node && ($node->key_cmp($key) < 0)) {
+    $node = $node->header()->[0];
+    $index++;
+  }
+
+  if ($node->key_cmp($key) == 0) {
+    $self->last_key( $node, $index );
+    return $index;
+  } else {
+    return;
+  }
+}
+
+
 sub _debug {
 
   my $self = shift;
@@ -994,7 +1321,7 @@ A detailed description of the methods used is below.
 
 =item new
 
-  $list = new SkipList();
+  $list = new List::SkipList();
 
 Creates a new skip list.
 
@@ -1002,33 +1329,34 @@ If you need to use a different L<node class|/"Node Methods"> for using
 customized L<comparison|/"key_cmp"> routines, you will need to specify a
 different class:
 
-  $list = new SkipList( node_class => 'MyNodeClass' );
+  $list = new List::SkipList( node_class => 'MyNodeClass' );
 
 See the L</"Customizing the Node Class"> section below.
 
 Specialized internal parameters may be configured:
 
-  $list = new SkipList( max_level => 32 );
+  $list = new List::SkipList( max_level => 32 );
 
-Defines a different maximum list level, or L</max_level>.  (The default
-is 32.) It is generally a good idea to leave this value alone unless
-you are using small lists.
+Defines a different maximum list level.
 
 The initial list (see the L</"list"> method) will be a
-L<random|/"_new_node_level"> number of levels, and will increase over time
-if inserted nodes have higher levels.
+L<random|/"_new_node_level"> number of levels, and will increase over
+time if inserted nodes have higher levels, up until L</max_level>
+levels.  See L</max_level> for more information on this parameter.
 
 You can also control the probability used to determine level sizes for
-each node by setting the L<P|/"p"> value:
+each node by setting the L<P|/"p"> and k values:
 
-  $list = new SkipList( p => 0.25 );
+  $list = new List::SkipList( p => 0.25, k => 1 );
 
-The value defaults to C<0.25>, which appears to be optimial.  Smaller
-values may reduce the average number of pointers per node but also
-reduce the efficiency of the algorithm.
+See  L<P|/p> for more information on this parameter.
 
-For more information on what these values mean, consult the references
-below in the L</"SEE ALSO"> section.
+You can enable duplicate keys by using the following:
+
+  $list = new List::SkipList( duplicates => 1 );
+
+This is an experimental feature. See the L</KNOWN ISSUES> section
+below.
 
 =item insert
 
@@ -1083,6 +1411,19 @@ the key cannot be found, returns C<undef>.
 This method is slightly faster than L</find_with_finger> since it does
 not return a search finger when called in list context.
 
+If you are searching for duplicate keys, you must use
+L</find_with_finger> or L</find_duplicates>.
+
+=item find_duplicates
+
+  @values = $list->find_duplicates( $key );
+
+  @values = $list->find_duplicates( $key, $finger );
+
+Returns an array of values from the list.
+
+This is an autoloading method.
+
 =item search
 
 Search is an alias to L</find>.
@@ -1124,6 +1465,8 @@ then the value of L</last_key> is assumed:
 
   $key = $list->next_key( $list->last_key );
 
+Note: calls to L</delete> will L</reset> the last key.
+
 =item next
 
   ($key, $value) = $list->next( $last_key, $finger );
@@ -1141,18 +1484,18 @@ This is an autoloading method.
   ($key, $finger, $value) = $list->last_key;
 
 Returns the last key or the last key and finger returned by a call to
-L</first_key> or L</next_key>.
+L</first_key>, L</next_key>, L</index_by_key>, L</key_by_index> or
+L</value_by_index>.  This is not the greatest key.
 
-Deletions and inserts will invalidate the L</last_key> value, although
-they may not L</reset> the last key.
+Deletions and inserts may invalidate the L</last_key> value.
+(Deletions will actually L</reset> the value.)
 
 Values for L</last_key> can also be set by including parameters,
 however this feature is meant for I<internal use only>:
 
-  $list->last_key( $key, $finger, $value );
+  $list->last_key( $node );
 
-No checking is done to make sure that the C<$key> and C<$value> pairs
-match, or that the C<$finger> is valid.
+Note that this is a change form versions prior to 0.71.
 
 =item reset
 
@@ -1160,6 +1503,35 @@ match, or that the C<$finger> is valid.
 
 Resets the L</last_key> to C<undef>. 
 
+=item index_by_key
+
+  $index = $list->index_by_key( $key );
+
+Returns the 0-based index of the key (as if the list were an array).
+I<This is not an efficient method of access.>
+
+This is an autoloading method.
+
+=item key_by_index
+
+  $key = $list->key_by_index( $index );
+
+Returns the key associated with an index (as if the list were an
+array).  Negative indices return the key from the end.  I<This is not
+an efficient method of access.>
+
+This is an autoloading method.
+
+=item value_by_index
+
+  $value = $list->value_by_index( $index );
+
+Returns the value associated with an index (as if the list were an
+array).  Negative indices return the value from the end.  I<This is not
+an efficient method of access.>
+
+This is an autoloading method.
+
 =item delete
 
   $value = $list->delete( $key );
@@ -1194,6 +1566,15 @@ Makes a copy of a list.  The L</"p">, L</"max_level"> and
 L<node class|/"_node_class"> are copied, although the exact structure of node
 levels is not copied.
 
+  $list2 = $list1->copy( $key_from, $finger, $key_to );
+
+Copy the list between C<$key_from> and C<$key_to> (inclusive).  If
+C<$finger> is defined, it will be used as a search finger to find
+C<$key_from>.  If C<$key_to> is not specified, then it will be assumed
+to be the end of the list.
+
+If C<$key_from> does not exist, C<undef> will be returned.
+
 This is an autoloading method.
 
 =item merge
@@ -1211,8 +1592,8 @@ This is an autoloading method.
 
   $list1->append( $list2 );
 
-Appends C<$list2> after C<$list1>.  The last key of C<$list1> must be less
-than the first key of C<$list2>.
+Appends (concatenates) C<$list2> after C<$list1>.  The last key of
+C<$list1> must be less than the first key of C<$list2>.
 
 Both lists should have the same L<node class|/"_node_class">.
 
@@ -1231,6 +1612,17 @@ or
 
 This is an autoloading method.
 
+=item truncate
+
+  $list2 = $list1->truncate( $key );
+
+Truncates C<$list1> and returns C<$list2> starting at C<$key>.
+Returns C<undef> is the key does not exist.
+
+It is asusmed that the key is not the first key in C<$list1>.
+
+This is an autoloading method.
+
 =item least
 
   ($key, $value) = $list->least;
@@ -1306,11 +1698,35 @@ Same as L</_search_with_finger>, only that a search finger is not returned.
 This is useful for searches where a finger is not needed.  The speed
 of searching is improved.
 
+=item k
+
+  $k = $list->k;
+
+Returns the I<k> value.
+
+  $list->k( $k );
+
+Sets the I<k> value.
+
+Higher values will on the average have less pointers per node, but
+take longer for searches.  See the section on the L<P|/p> value.
+
 =item p
 
   $plevel = $list->p;
 
-Returns the I<P> value.  Intended for internal use only.
+Returns the I<P> value.
+
+  $list->p( $plevel );
+
+Changes the value of I<P>.  Lower values will on the average have less
+pointers per node, but will take longer for searches.
+
+The probability that a particular node will have a forward pointer at
+level I<i> is: I<p**(i+k-1)>.
+
+For more information, consult the references below in the
+L</"SEE ALSO"> section.
 
 =item max_level
 
@@ -1318,6 +1734,23 @@ Returns the I<P> value.  Intended for internal use only.
 
 Returns the maximum level that L</_new_node_level> can generate.
 
+  eval {
+    $list->max_level( $level );
+  };
+
+Changes the maximum level.  If level is less than L</MIN_LEVEL>, or
+greater than L</MAX_LEVEL> or the current list L</level>, this will fail
+(hence the need for setting it in an C<eval> block).
+
+The value defaults to L</MAX_LEVEL>, which is 32.  There is usually no
+need to change this value, since the maximum level that a new node
+will have will not be greater than it actually needs, up until 2^32
+nodes.  (The current version of this module is not designed to handle
+lists larger than 2^32 nodes.)
+
+Decreasing the maximum level to less than is needed will likely
+degrade performance.
+
 =item _new_node_level
 
   $level = $list->_new_node_level;
@@ -1342,22 +1775,20 @@ C<List::SkipList::Node> (See L<below|/"Node Methods">.)
 
 The key and value for this node are undefined.
 
-=item level
-
-  $level = $list->level;
-
-Returns the number of levels in the list.  It is the same as
-
-  $level = $list->list->level;
-
 =item _first_node
 
-  ($node, $finger) = _first_node;
+  $node = $list->_first_node;
 
-Returns the first node with a key (the second node) in a list and the
-finger.  This is used by the L</merge> method.
+Returns the first node with a key (the second node) in a list.  This
+is used by the L</first_key>, L</least>, L</append> and L</merge>
+methods.
 
-This is an autoloading method.
+=item _greatest_node
+
+  $node = $list->_greatest_node;
+
+Returns the last node in the list.  This is used by the L</append> and
+L</greatest> methods.
 
 =item _node_class
 
@@ -1366,14 +1797,22 @@ This is an autoloading method.
 Returns the name of the node class used.  By default this is the
 C<List::SkipList::Node>, which is discussed below.
 
+=item _build_distribution
+
+  $list->_build_distribution;
+
+Rebuilds the probability distribution array C<{P_LEVELS}> upon calls
+to L</_set_p> and L</_set_k>.
+
 =item _set_node_class
 
 =item _set_max_level
 
 =item _set_p
 
-These methods are used only during initialization of the object.
-I<Do not call these methods after the object has been created!>
+=item _set_k
+
+These methods are used during initialization of the object.
 
 =item _debug
 
@@ -1404,11 +1843,10 @@ convention.
 
   $key = $node->key;
 
-Returns the node's key.
+Returns the key.
 
-  $node->key( $key );
-
-When used with an argument, sets the node's key.
+Note that as of version 0.70, this method is read-only.  We should not
+change the key once a node has been added to the list.
 
 =item key_cmp
 
@@ -1462,13 +1900,9 @@ Returns the forward list array of the node. This is an array of nodes
 which point to the node returned, where each index in the array refers
 to the level.
 
-  $node->header( $header_ref );
-
-When used with an argument, sets the forward list.  It does not check
-if list elements are of the correct type.
-
-Note that the interface has changed. This method only accepts or
-returns header references.
+Note that as of version 0.70, this method is read-only.  Since it only
+returns header references (as of version 0.50), that reference can be
+used to modify forward pointers.
 
 =item level
 
@@ -1591,37 +2025,47 @@ flexibility in defining a custom node class.
 
 See the included F<Benchmark.txt> file for performance comparisons.
 
-=head1 TODO
-
-The following features may be added in future versions:
+=head1 KNOWN ISSUES
 
 =over
 
-=item Accessing list nodes by index number as well as key
+=item Undefined Values
 
-The ability to tie a list to an array as well as a hash, probably as a
-subclass since to implement it efficiently would require some extra
-bookkeeping.
+Certain methods such as L</find> and L</delete> will return the the
+value associated with a key, or C<undef> if the key does not exist.
+However, if the value is C<undef>, then these functions will appear to
+claim that the key cannot be found.
 
-=item Splitting lists
+In such circumstances, use the L</exists> method to test for the
+existence of a key.
 
-The ability to split a list into multiple segments.
+=item Duplicate Keys
 
-=item Deterministic Skip Lists
+Duplicate keys are an experimental feature in this module, since most
+methods have been designed for unique keys only.
 
-An additional module (probably a subclass of List::SkipList) to
-implement deterministic skip lists (DSLs), probably as a 1-2-3 skip
-list.
+Access to duplicate keys is akin to a stack.  When a duplicate key is
+added, it is always inserted I<before> matching keys.  In searches, to
+find duplicate keys one must use L</find_with_finger> or the
+L</find_duplicates> method.
 
-=back
+The L</copy> method will reverse the order of duplicates.
 
-=head1 CAVEATS
+The behavior of the L</merge> and L</append> methods is not defined
+for duplicates.
+
+=item Non-Determinism
 
 Skip lists are non-deterministic.  Because of this, bugs in programs
 that use this module may be subtle and difficult to reproduce without
 many repeated attempts.  This is especially true if there are bugs in
 a L<custom node|/"Customizing the Node Class">.
 
+=back
+
+Additional issues may be listed on the CPAN Request Tracker at
+L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=List-SkipList>.
+
 =head1 AUTHOR
 
 Robert Rothenberg <rrwo at cpan.org>
